/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include "c_fun_args.h"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "list.h"
#include "string.h"

static char *buffer;
static int size;
static int maxsize;

static char stack[128];
static int stack_size;

/* -1 error 0 ok */
int push_or_pop(char s)
{
  switch(s) {
    case ')' :
      if (stack_size == 0) goto pop_error;
      stack_size--;
      if (stack[stack_size] != '(') goto pop_error;
      break;
    case ']' :
      if (stack_size == 0) goto pop_error;
      stack_size--;
      if (stack[stack_size] != '[') goto pop_error;
      break;
    case '}' :
      if (stack_size == 0) goto pop_error;
      stack_size--;
      if (stack[stack_size] != '{') goto pop_error;
      break;
    default:
      if (s != '(' && s != '[' && s != '{') return 0;
      if (stack_size == 128) {
        fprintf(stderr, "parenthesis too much nested\n");
        return -1;
      }
      stack[stack_size] = s;
      stack_size++;
      break;
  }
  return 0;

pop_error:
  fprintf(stderr, "wrong parenthesis\n");
  return -1;
}

static char *p;

void remove_white_spaces(void)
{
  while (*p && isspace(*p)) p++;
}

/* -1 error 0 ok */
int parse_struct_union_enum(void)
{
  int has_identifier = 0;

  remove_white_spaces();
  if (!*p) return -1;

  /* optional identifier */
  if (*p == '_' || isalpha(*p)) {
    has_identifier = 1;
    while (*p && (*p == '_' || isalnum(*p))) p++;
  }

  remove_white_spaces();
  if (!*p) return -1;

  if (*p != '{' && !has_identifier) goto sue_error;

  if (push_or_pop(*p) == -1) return -1;
  while (*p && stack_size) {
    p++;
    if (push_or_pop(*p) == -1) return -1;
  }
  if (!*p) goto sue_error;
  p++;

  return 0;

sue_error:
  fprintf(stderr, "bad union/struct/enum declaration\n");
  return -1;
}

/* -1 error, arg name ok */
string parse_argument(void)
{
  char *id;
  char c;
  string s;

  /* we accept more than what the spec (c99) says, but anyway the C compiler
   * will complain later on, so no problem
   */
  int type_specified = 0;

  /* first stuff: type specifier */
  /* barbarian code */
  while (1) {
    remove_white_spaces();
    if (!*p) goto arg_error;

    if (!strncmp(p, "typedef",  7)) { p+=7; continue; }
    if (!strncmp(p, "extern",   6)) { p+=6; continue; }
    if (!strncmp(p, "static",   6)) { p+=6; continue; }
    if (!strncmp(p, "auto",     4)) { p+=4; continue; }
    if (!strncmp(p, "register", 8)) { p+=8; continue; }

    if (!strncmp(p, "const",    5)) { p+=5; continue; }
    if (!strncmp(p, "restrict", 8)) { p+=8; continue; }
    if (!strncmp(p, "volatile", 8)) { p+=8; continue; }

    if (!strncmp(p, "inline",   6)) { p+=6; continue; }

    if (!strncmp(p, "void",     4)) { p+=4; type_specified = 1; continue; }
    if (!strncmp(p, "char",     4)) { p+=4; type_specified = 1; continue; }
    if (!strncmp(p, "short",    5)) { p+=5; type_specified = 1; continue; }
    if (!strncmp(p, "int",      3)) { p+=3; type_specified = 1; continue; }
    if (!strncmp(p, "long",     4)) { p+=4; type_specified = 1; continue; }
    if (!strncmp(p, "float",    5)) { p+=5; type_specified = 1; continue; }
    if (!strncmp(p, "double",   6)) { p+=6; type_specified = 1; continue; }
    if (!strncmp(p, "signed",   6)) { p+=6; type_specified = 1; continue; }
    if (!strncmp(p, "unsigned", 8)) { p+=8; type_specified = 1; continue; }
    if (!strncmp(p, "_Bool",    5)) { p+=5; type_specified = 1; continue; }
    if (!strncmp(p, "_Complex", 8)) { p+=8; type_specified = 1; continue; }
    if (!strncmp(p, "_Imaginary", 10)) { p+=10; type_specified = 1; continue; }

    if (!strncmp(p, "struct", 6) || !strncmp(p, "union", 5) ||
        !strncmp(p, "enum", 4)) {
      if (*p == 's') p+=6;
      else if (*p == 'u') p+=5;
      else p += 4;
      if (parse_struct_union_enum() == -1) return -1;
      type_specified = 1;
      continue;
    }

    /* might be a user defined type name */
    if (*p == '_' || isalpha(*p)) {
      /* if type_specified is true this identifier must be understood as the
       * name of the argument
       */
      if (type_specified) break;
      while (*p && (*p == '_' || isalnum(*p))) p++;
      type_specified = 1;
      continue;
    }

    /* might be the beginning of a declarator */
    if (*p == '*' || *p == '(') break;

    goto arg_error;
  }

  /* now: declarator */
  while (1) {
    remove_white_spaces();
    if (!*p) goto arg_error;
    if (push_or_pop(*p) == -1) return -1;
    if (*p == '_' || isalpha(*p)) break;
    if (*p != '*' && *p !='(') goto arg_error;
    p++;
  }

  /* we've got the identifier, let's save it */
  id = p;
  while (*p && (*p == '_' || isalnum(*p))) p++;
  c = *p;
  *p = 0;
  s = new_string(id);
  if (s == -1) return -1;
  *p = c;

  /* now, let's eat up to the end of the declaration */
  while (*p && (stack_size || *p != ',')) {
    if (push_or_pop(*p) == -1) return -1;
    p++;
  }
  if (stack_size) goto arg_error;

#if 0
  /* let's replace the ',' by a ';' */
  if (*p) { *p = ';'; p++; }
#endif
  if (*p) p++;

  return s;

arg_error:
  fprintf(stderr, "bad arguments' declaration\n");
  return -1;
}

/* -1 error, list if ok */
/* list is: <arg 1 name> <arg 2 name> ... <arg n name> <arg string> */
list parse_arguments(char *sn)
{
  list l = new_list();
  string s;
  if (l == -1) return -1;

  p = sn;
  remove_white_spaces();
  while (*p) {
    string sa = parse_argument();
    if (sa == -1) return -1;
    if (list_add_int(l, sa) == -1) return -1;
    remove_white_spaces();
  }

#if 0
  s = new_string_concat(sn, ";");
#endif
  s = new_string(sn);
  if (s == -1 || list_add_int(l, s) == -1) return -1;

  return l;
}

void close_arguments(void)
{
  if (!maxsize) return;

  free(buffer);
  size = maxsize = 0;
  buffer = 0;
}

#if 0
int main(void)
{
  char p[] = "int x, char *y, void f(int, char *)";
  list l = parse_arguments(p);
  if (l == -1) fprintf(stderr, "roro\n");
  return 0;
}
#endif
