/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include "files.h"

#include <string.h>

#include "convert.h"

static FILE *input_file;
static FILE *output_h;
static FILE *output_c;

static char *file_name;

char *hout;

/* -1 error, 0 ok */
int init_files(char *v)
{
  int l = strlen(v);
  if (l < 2) goto bad_name;
  if (v[l-1] != 'k' || v[l-2] != '.') goto bad_name;

  input_file = fopen(v, "r");
  if (!input_file) {
    perror(v);
    return -1;
  }

  v[l-1] = 'c';
  output_c = fopen(v, "w");
  if (!output_c) {
    perror(v);
    return -1;
  }

  if (hout) {
    output_h = fopen(hout, "w");
  } else {
    v[l-1] = 'h';
    output_h = fopen(v, "w");
  }
  if (!output_h) {
    perror(v);
    return -1;
  }

  v[l-1] = 'k';
  file_name = v;

  return 0;

bad_name:
  fprintf(stderr, "%s: wrong input file (extension must be .k)\n", v);
  return -1;
}

/* -1 error, 0 ok */
int close_file(FILE **f)
{
  int ret;
  if (!*f) return 0;
  ret = fclose(*f);
  if (ret) fprintf(stderr, "close error\n");
  *f = 0;
  return ret;
}

/* -1 error, 0 ok */
int close_files(void)
{
  if (close_file(&input_file) == 1) goto close_error;
  if (close_file(&output_c) == 1) goto close_error;
  if (close_file(&output_h) == 1) goto close_error;
  return 0;
close_error:
  close_file(&input_file);
  close_file(&output_c);
  close_file(&output_h);
  return -1;
}

int get(void)
{
  return getc(input_file);
}

void unget(int c)
{
  ungetc(c, input_file);
}

int put_c(char *s)
{
  char *s1 = convert(s);
  if (!s1) return -1;
  if (fprintf(output_c, "%s", s1) != -1) return 0;
  fprintf(stderr,
          "error writing in the C output file. No more space on disk?\n");
  return -1;
}

int put_h(char *s)
{
  char *s1 = convert(s);
  if (!s1) return -1;
  if (fprintf(output_h, "%s", s1) != -1) return 0;
  fprintf(stderr,
          "error writing in the H output file. No more space on disk?\n");
  return -1;
}

char *get_current_file(void)
{
  return file_name;
}
