/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include "generate.h"

#include <string.h>
#include <ctype.h>

#include "widget.h"
#include "files.h"
#include "list.h"
#include "gtk_widget.h"
#include "globals.h"

extern list cinclude;
extern list hinclude;
extern list ccode;
extern list hcode;

static char *header =
    "/* generated by gtkcooker (http://www.nongnu.org/gtkcooker) */\n\n";

/* -1 error, 0 ok */
int declare_signals(gtk_widget gw)
{
  list l = WIDGET_SIGNALS(gw);
  keyword k;
  int i;

  for (i=0; i<list_size(l); i+=3) {
    k = (keyword)list_get_int(l, i);
    switch (k) {
      default:
        PUT_H("  unsigned long ");
        PUT_H(get_identifier(WIDGET_NAME(gw)));
        PUT_H("_");
        PUT_H(get_keyword(k));
        PUT_H(";\n");
        break;
      case KEYWORD_POINTER:
        PUT_H("  unsigned long ");
        PUT_H(get_identifier(WIDGET_NAME(gw)));
        PUT_H("_pointer_motion");
        PUT_H(";\n");
        PUT_H("  unsigned long ");
        PUT_H(get_identifier(WIDGET_NAME(gw)));
        PUT_H("_pointer_press");
        PUT_H(";\n");
        PUT_H("  unsigned long ");
        PUT_H(get_identifier(WIDGET_NAME(gw)));
        PUT_H("_pointer_release");
        PUT_H(";\n");
        break;
    }
//    PUT_H(get_identifier((identifier)list_get_int(l, i+1)));
  }

  return 0;
}

/* -1 error, 0 ok */
int create_struct(widget w)
{
  char *vars;
  int i;
  int n;

  PUT_H("typedef struct ");
  PUT_H(widget_name(w));
  PUT_H(" {\n");

  /* the variables of the widget */
  vars = widget_vars(w);
  if (vars) { PUT_H(vars); PUT_H("\n"); }

  /* the stuff needed by every gtk_widget */
  n = widget_nb_gtk_widgets(w);
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_struct)
      if (driver->generate_struct(gw) == -1) return -1;
    if (declare_signals(gw) == -1) return -1;
  }

  PUT_H("} ");
  PUT_H(widget_name(w));
  PUT_H(";\n\n");

  return 0;
}

/* -1 error, 0 ok */
int create_init_code(widget w)
{
  list l = widget_arguments(w);
  int i;
  int n = list_size(l);

  /* in the .h put the declarator */
  PUT_H("int init_");
  PUT_H(widget_name(w));
  PUT_H("(");
  PUT_H(widget_name(w));
  PUT_H(" *this");
  if (n) {
    PUT_H(", ");
    PUT_H(get_string((string)list_get_int(l, n-1)));
  }
  PUT_H(");\n");

  /* in the .C, defines the signal handlers */
  n = widget_nb_gtk_widgets(w);
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_signal_handlers)
      if (driver->generate_signal_handlers(w, gw) == -1) return -1;
  }

  /* in the .C, defines the init function */
  n = list_size(l);
  PUT_C("/* returns -1 on error, 0 if everything is fine */\n");
  PUT_C("int init_");
  PUT_C(widget_name(w));
  PUT_C("(");
  PUT_C(widget_name(w));
  PUT_C(" *this");
  if (n) {
    PUT_C(", ");
    PUT_C(get_string((string)list_get_int(l, n-1)));
  }
  PUT_C(")\n{\n");

  /* put the local widget stuff */
  n = widget_nb_gtk_widgets(w);
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_variable_declaration)
      if (driver->generate_variable_declaration(gw) == -1) return -1;
  }
  if (n) PUT_C("\n");

  /* put the init code (if there is one) */
  if (widget_init_code(w)) {
    PUT_C("  ");
    PUT_C(widget_init_code(w));
    PUT_C("\n\n");
  }

  /* create the widgets */
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_variable_instanciation)
      if (driver->generate_variable_instanciation(gw) == -1) return -1;
  }
  if (n) PUT_C("\n");

  /* pack the widgets the ones in the others */
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_pack)
      if (driver->generate_pack(gw) == -1) return -1;
  }
  if (n) PUT_C("\n");

  /* after-pack the widgets */
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_post_pack)
      if (driver->generate_post_pack(gw) == -1) return -1;
  }
  if (n) PUT_C("\n");

  /* the signals */
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->generate_signals)
      if (driver->generate_signals(gw) == -1) return -1;
  }

  /* put the post init code (if there is one) */
  if (widget_post_init_code(w)) {
    PUT_C("  ");
    PUT_C(widget_post_init_code(w));
    PUT_C("\n\n");
  }

  /* ok */
  PUT_C("  return 0;\n}\n\n");

  return 0;
}

/* -1 error, 0 ok */
int put_filename_for_ifdef(char *file_name)
{
  int i;
  char c[2];
  c[1] = 0;

  PUT_H("_");
  for (i=0; file_name[i]; i++) {
    if (isalpha(file_name[i])) {
      /* we force last char to be H, it is 'k' at first, we want h */
      if (file_name[i+1] == 0) c[0] = 'H';
      else c[0] = toupper(file_name[i]);
      PUT_H(c);
    } else PUT_H("_");
  }
  PUT_H("_");

  return 0;
}

/* -1 error, 0 ok */
int generate(char *file_name)
{
  int i;
  int n;

  /* the header */
  PUT_H(header);
  PUT_C(header);

  /* the ifndef */
  PUT_H("#ifndef ");
  if (put_filename_for_ifdef(file_name) == -1) return -1;
  PUT_H("\n#define ");
  if (put_filename_for_ifdef(file_name) == -1) return -1;
  PUT_H("\n\n");

  /* some includes */
  if (hide_gtk)
    PUT_C("#include <gtk/gtk.h>\n");
  else
    PUT_H("#include <gtk/gtk.h>\n");
  PUT_C("#include \"");
  file_name[strlen(file_name)-1] = 'h';
  PUT_C(file_name);
  PUT_C("\"\n\n");

  /* the include */
  n = list_size(hinclude);
  for (i=0; i<n; i++) {
    PUT_H("#include");
    PUT_H(get_string((string)list_get_int(hinclude, i)));
    PUT_H("\n");
  }
  PUT_H("\n");

  n = list_size(cinclude);
  for (i=0; i<n; i++) {
    PUT_C("#include");
    PUT_C(get_string((string)list_get_int(cinclude, i)));
    PUT_C("\n");
  }
  PUT_C("\n");

  /* the 'c' and 'h' constructs */
  n = list_size(hcode);
  for (i=0; i<n; i++) {
    PUT_H(get_string((string)list_get_int(hcode, i)));
    PUT_H("\n");
  }

  n = list_size(ccode);
  for (i=0; i<n; i++) {
    PUT_C(get_string((string)list_get_int(ccode, i)));
    PUT_C("\n");
  }

  /* in the .h put the structures */
  n = nb_widgets();
  for (i=0; i<n; i++)
    if (create_struct(i) == -1) return -1;

  /* generate init code */
  n = nb_widgets();
  for (i=0; i<n; i++)
    if (create_init_code(i) == -1) return -1;

  /* endif in the .h */
  PUT_H("\n#endif /* ");
  if (put_filename_for_ifdef(file_name) == -1) return -1;
  PUT_H(" */\n");

  return 0;
}
