/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

%{

/* C declarations */

/* the mode: the lexer should behave in various ways, depending
 * on this variable
 * possible values:
 * LEXER_NORMAL		parse a word or a symbol, one at a time
 * LEXER_C_CODE		parse some code between '{' and '}'
 * LEXER_F_ARGS		parse a function arguments' list (something like
                        "int x, char* y") or a function call's parameters
			(like "x, y, "a", 'b', 2")
 * LEXER_C_STRING	parse a C string (the ending '"' won't be eaten)
 * LEXER_INCLUDE_STUFF	parse between #include and End Of Line
 * LEXER_INIT_VARS	parse a C block, but omit the '{' and '}'
 *			(the C block shall contain only variables'
 *			declarations)
 * LEXER_C_CODE_NO_BRACKET same meaning as LEXER_INIT_VARS
 */
#define LEXER_NORMAL			0
#define LEXER_C_CODE			1
#define LEXER_F_ARGS			2
#define LEXER_C_STRING			3
#define LEXER_INCLUDE_STUFF		4
#define LEXER_INIT_VARS			5
#define LEXER_C_CODE_NO_BRACKET		6

#include <string.h>

#include "widget.h"
#include "string.h"
#include "identifier.h"
#include "list.h"
#include "c_fun_args.h"
#include "files.h"
#include "keywords.h"

static int lexer_mode = LEXER_NORMAL;
static int line = 0;
static int init_first_shot;

void yyerror(char *);
static char *buffer;
static int buffer_size;
static int buffer_maxsize;
static int put(int);
void end_parse(void);
static int yylex(void);

static keyword buffer_keyword;

widget current_widget;

extern list cinclude;
extern list hinclude;
extern list ccode;
extern list hcode;

%}

/* yacc definitions */

%start file

%token WIDGET IDENTIFIER FLOAT INTEGER C_STRING C_CODE F_ARGS INCLUDE_STUFF
%token POST_INIT INIT INCLUDE CINCLUDE HINCLUDE INIT_VARS CCODE HCODE KEYWORD

%union {
  widget widget;
  int integer;
  property_value property_value;
  widget_argument_value widget_argument_value;
  float f;
}

%type <integer> widget_declaration_arguments
%type <integer> widget_arguments
%type <integer> widget_type
%type <integer> widget_pack_name
%type <integer> widget_name
%type <integer> widget_pack_name_list
%type <integer> keyword_value
%type <integer> identifier_value
%type <f> float_value
%type <f> float_or_integer_value
%type <integer> integer_value
%type <integer> string_value
%type <integer> function_call_value
%type <integer> function_name
%type <integer> function_arguments
%type <integer> c_code_value
%type <integer> include_string
%type <integer> identifier
%type <integer> keyword
%type <integer> property_name
%type <property_value> property_value
%type <widget_argument_value> widget_argument
%type <integer> widget_packing_attributes
%type <integer> widget_packing_attributes_list

%%

/* the grammar */

file:
	  includes widget_list
	;

includes:
	  /* nothing */
	  {}
	| includes '#' include;
	;

include:
	  INCLUDE include_string
	  {
	    if (list_add_int(cinclude, $2) == -1) YYERROR;
	    if (list_add_int(hinclude, $2) == -1) YYERROR;
	  }
	| CINCLUDE include_string
	  {
	    if (list_add_int(cinclude, $2) == -1) YYERROR;
	  }
	| HINCLUDE include_string
	  {
	    if (list_add_int(hinclude, $2) == -1) YYERROR;
	  }
	;

widget_list:
	  /* nothing */
	  {}
	| widget_list widget
	  {}
	| widget_list code
	  {}
	;

code:
	  CCODE '{'
	  { lexer_mode = LEXER_C_CODE_NO_BRACKET; }
	  C_CODE
	  { lexer_mode = LEXER_NORMAL; }
	  '}'
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    if (list_add_int(ccode, s) == -1) YYERROR;
	  }
	| HCODE '{'
	  { lexer_mode = LEXER_C_CODE_NO_BRACKET; }
	  C_CODE
	  { lexer_mode = LEXER_NORMAL; }
	  '}'
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    if (list_add_int(hcode, s) == -1) YYERROR;
	  }
	;

widget:
	  widget_declaration '{' widget_body '}'
	  {}
	;

widget_declaration:
	  WIDGET widget_declaration_simple
	  {
	    init_first_shot = 0;
	  }
	| WIDGET widget_declaration_complex
	  {
	    init_first_shot = 0;
	  }
	;

widget_declaration_simple:
	  identifier
	  {
	    current_widget = new_widget($1, -1);
	    if (current_widget == -1) YYERROR;
	  }
	| keyword
	  {
	    fprintf(stderr, "%s:%d: keyword '%s' used as widget name, error\n",
	            get_current_file(), line+1, get_keyword($1));
	    YYERROR;
	  }
	;

widget_declaration_complex:
	  identifier '(' widget_declaration_arguments ')'
	  {
	    current_widget = new_widget($1, $3);
	    if (current_widget == -1) YYERROR;
	  }
	;

widget_declaration_arguments:
	  { lexer_mode = LEXER_F_ARGS; }
	  F_ARGS
	  {
	    list l;
	    lexer_mode = LEXER_NORMAL;
	    l = parse_arguments(buffer);
	    if (l == -1) YYERROR;
	    $$ = l;
	  }

widget_body:
	  init_vars init_code widget_body_next post_init_code
	  {}
	;

init_vars:
	  /* nothing */
	  {}
	| '{'
	  { lexer_mode = LEXER_INIT_VARS; }
	  INIT_VARS
	  { lexer_mode = LEXER_NORMAL; }
	  '}'
	  {
	    if (set_init_vars(current_widget, buffer) == -1) YYERROR;
	  }
	;

init_code:
	  { init_first_shot = 1; }
	| INIT c_code_value
	  {
	    if (set_init_code(current_widget, $2) == -1) YYERROR;
	    init_first_shot = 1;
	  }
	;

post_init_code:
	  /* nothing */
	  {}
	| POST_INIT c_code_value
	  {
	    if (set_post_init_code(current_widget, $2) == -1) YYERROR;
	  }
	;

widget_body_next:
	  /* nothing */
	  {}
	| widget_body_next widget_unit_declaration
	  {}
	;

widget_unit_declaration:
	  widget_element_declaration
	  {}
	| widget_property_setting
	  {}
	| widget_packing
	  {}
	;

widget_element_declaration:
	  widget_type widget_name ';'
	  {
	    /* names of elements must be unique */
	    if (widget_find_element(current_widget, $2) != -1) {
	      fprintf(stderr, "element %s declared twice\n",
		      get_identifier($2));
	      YYERROR;
	    }
	    if (widget_add_element(current_widget, $1, $2, -1) == -1)
	      YYERROR;
	  }
	| widget_type widget_name '(' widget_arguments ')' ';'
	  {
	    /* names of elements must be unique */
	    if (widget_find_element(current_widget, $2) != -1) {
	      fprintf(stderr, "element %s declared twice\n",
		      get_identifier($2));
	      YYERROR;
	    }
	    if (widget_add_element(current_widget, $1, $2, $4) == -1)
	      YYERROR;
	  }
	;

widget_property_setting:
	  widget_name '.' property_name '=' property_value ';'
	  {
	    widget_element e = widget_find_element(current_widget, $1);
	    if (e == -1) {
	      fprintf(stderr, "element %s not declared\n",
		      get_identifier($1));
	      YYERROR;
	    }
	    if (set_property(current_widget, e, $3, $5) == -1) YYERROR;
	  }
	| widget_name '[' integer_value ']' '.' property_name '='
	  property_value ';'
	  {
	    widget_element e = widget_find_element(current_widget, $1);
	    if (e == -1) {
	      fprintf(stderr, "element %s not declared\n",
		      get_identifier($1));
	      YYERROR;
	    }
	    if ($3 < 0 || $3 >= list_size(widget_element_arguments(
                                              current_widget, e)) / 2) {
	      fprintf(stderr,
                      "number %d unallowed in [n] notation of element %s\n",
		      $3, get_identifier($1));
	      fprintf(stderr, "max value allowed is %d\n",
	        list_size(widget_element_arguments(current_widget, e))/2-1);
	      YYERROR;
	    }
	    if (set_argument_property(current_widget, e, $3, $6, $8) == -1)
	      YYERROR;
	  }
	;

widget_packing:
	  widget_name '>' widget_packing_attributes widget_pack_name_list ';'
	  {
	    int i;
	    int l;
	    widget_element e = widget_find_element(current_widget, $1);
	    if (e == -1) {
	      fprintf(stderr, "element %s not declared\n",
		      get_identifier($1));
	      YYERROR;
	    }
	    /* menu may have "0" (represented as -2) as child */
	    l = list_size($4);
	    if (widget_element_type(current_widget, e) != KEYWORD_MENU) {
	      for (i=0; i<l; i++)
	        if (list_get_int($4, i) == -2) {
	          fprintf(stderr,
	              "packing 0 not allowed for element %s (type %s)\n",
	              get_identifier(widget_element_name(current_widget, e)),
	              get_keyword(widget_element_type(current_widget, e)));
	          YYERROR;
	        }
	    }
	    if (pack_new_row(current_widget, e, $3) == -1) YYERROR;
	    if (pack_append(current_widget, e, $4) == -1) YYERROR;
	    /* set dad */
	    for (i=0; i<l; i++) {
	      widget_element x = (widget_element)list_get_int($4, i);
	      if (x == -2 || x == -3 || x == -4) continue;
	      widget_element_set_dad(current_widget, x, e);
	    }
	  }
	| widget_name '+' widget_pack_name_list ';'
	  {
	    int i;
	    int l;
	    widget_element e = widget_find_element(current_widget, $1);
	    if (e == -1) {
	      fprintf(stderr, "element %s not declared\n",
		      get_identifier($1));
	      YYERROR;
	    }
	    /* menu may have "0" (represented as -2) as child */
	    l = list_size($3);
	    if (widget_element_type(current_widget, e) != KEYWORD_MENU) {
	      for (i=0; i<l; i++)
	        if (list_get_int($3, i) == -2) {
	          fprintf(stderr,
	              "packing 0 not allowed for element %s (type %s)\n",
	              get_identifier(widget_element_name(current_widget, e)),
	              get_keyword(widget_element_type(current_widget, e)));
	          YYERROR;
	        }
	    }
	    if (pack_append(current_widget, e, $3) == -1) YYERROR;
	    /* set dad */
	    for (i=0; i<l; i++) {
	      widget_element x = (widget_element)list_get_int($3, i);
	      if (x == -2 || x == -3 || x == -4) continue;
	      widget_element_set_dad(current_widget, x, e);
	    }
	  }
	;

widget_packing_attributes:
	  /* nothing */
	  {
	    $$ = -1;
	  }
	| '(' widget_packing_attributes_list ')'
	  {
	    $$ = $2;
	  }
	;

widget_packing_attributes_list:
	  keyword ':' property_value
	  {
	    list l = new_list();
	    if (l == -1) YYERROR;
	    if (list_add_int(l, $1) == -1) YYERROR;
	    if (list_add_int(l, $3.type) == -1) YYERROR;
	    if ($3.type == PV_NUMBER) {
	      if (list_add_float(l, $3.v.d) == -1) YYERROR;
	    } else {
	      if (list_add_int(l, $3.v.value) == -1) YYERROR;
	    }
	    $$ = l;
	  }
	| widget_packing_attributes_list ',' keyword ':' property_value
	  {
	    if (list_add_int($1, $3) == -1) YYERROR;
	    if (list_add_int($1, $5.type) == -1) YYERROR;
	    if ($5.type == PV_NUMBER) {
	      if (list_add_float($1, $5.v.d) == -1) YYERROR;
	    } else {
	      if (list_add_int($1, $5.v.value) == -1) YYERROR;
	    }
	    $$ = $1;
	  }
	;

widget_pack_name_list:
	  widget_pack_name
	  {
	    list l = new_list();
	    if (l == -1) YYERROR;
	    if (list_add_int(l, $1) == -1) YYERROR;
	    $$ = l;
	  }
	| widget_pack_name_list ',' widget_pack_name
	  {
	    if (list_add_int($1, $3) == -1) YYERROR;
	    $$ = $1;
	  }
	;

widget_pack_name:
	  identifier
	  {
	    widget_element e = widget_find_element(current_widget, $1);
	    if (e == -1) {
	      fprintf(stderr, "packing unknown widget %s\n",
	        get_identifier($1));
	      YYERROR;
	    }
	    $$ = e;
	  }
	| keyword
	  {
	    /* special case of "separator", we return -3 */
	    if ($1 == KEYWORD_SEPARATOR)
	      $$ = -3;
	    /* special case of "resizer", we return -4 */
	    else if ($1 == KEYWORD_RESIZER)
	      $$ = -4;
	    else YYERROR;
	  }
	| integer_value
	  {
	    if ($1 != 0) {
	      fprintf(stderr, "packing a number (%d) is not allowed\n", $1);
	      YYERROR;
	    }
	    $$ = -2;
	  }
	;

widget_type:
	  keyword
	  { $$ = $1; }
	;

widget_name:
	  identifier
	  { $$ = $1; }
	| keyword
	  {
	    fprintf(stderr, "%s:%d: keyword '%s' used as widget name, error\n",
	            get_current_file(), line+1, get_keyword($1));
	    YYERROR;
	  }
	;

widget_arguments:
	  widget_argument
	  {
	    int l = new_list();
	    if (l == -1) YYERROR;
	    if (list_add_int(l, $1.type) == -1) YYERROR;
	    if ($1.type == WA_NUMBER) {
	      if (list_add_float(l, $1.v.d) == -1) YYERROR;
	    } else {
	      if (list_add_int(l, $1.v.value) == -1) YYERROR;
	    }
	    $$ = l;
	  }
	| widget_arguments ',' widget_argument
	  {
	    if (list_add_int($1, $3.type) == -1) YYERROR;
	    if ($3.type == WA_NUMBER) {
	      if (list_add_float($1, $3.v.d) == -1) YYERROR;
	    } else {
	      if (list_add_int($1, $3.v.value) == -1) YYERROR;
	    }
	    $$ = $1;
	  }
	;

widget_argument:
	  keyword_value
	  {
	    $$.type = WA_KEYWORD;
	    $$.v.value = $1;
	  }
	| identifier_value
	  {
	    $$.type = WA_IDENTIFIER;
	    $$.v.value = $1;
	  }
	| float_or_integer_value
	  {
	    $$.type = WA_NUMBER;
	    $$.v.d = $1;
	  }
	| string_value
	  {
	    $$.type = WA_STRING;
	    $$.v.value = $1;
	  }
	;

property_name:
	  keyword
	  { $$ = $1; }
	| identifier
	  {
	    fprintf(stderr, "%s:%d: property '%s' does not exist, error\n",
	            get_current_file(), line+1, get_identifier($1));
	    YYERROR;
	  }
	;

property_value:
	  keyword_value
	  {
	    $$.type = PV_KEYWORD;
	    $$.v.value = $1;
	  }
	| identifier_value
	  {
	    $$.type = PV_IDENTIFIER;
	    $$.v.value = $1;
	  }
	| float_or_integer_value
	  {
	    $$.type = PV_NUMBER;
	    $$.v.d = $1;
	  }
	| string_value
	  {
	    $$.type = PV_STRING;
	    $$.v.value = $1;
	  }
	| function_call_value
	  {
	    //if (function_call_to_c_code() == -1) YYERROR;
	    $$.type = PV_C_CODE;
	    $$.v.value = $1;
	  }
	| c_code_value
	  {
	    $$.type = PV_C_CODE;
	    $$.v.value = $1;
	  }
	;

keyword_value:
	  keyword
	  { $$ = $1; }
	;

identifier_value:
	  identifier
	  { $$ = $1; }
	;

float_or_integer_value:
	  integer_value
	  {
	    $$ = $1;
	  }
	| float_value
	  {
	    $$ = $1;
	  }
	;

integer_value:
	  INTEGER
	  {
	    int x;
	    if (sscanf(buffer, "%d", &x) != 1) {
	      fprintf(stderr, "bad number\n");
	      YYERROR;
	    }
	    $$ = x;
	  }
	;

float_value:
	  FLOAT
	  {
	    float x;
	    if (sscanf(buffer, "%f", &x) != 1) {
	      fprintf(stderr, "bad number\n");
	      YYERROR;
	    }
	    $$ = x;
	  }
	;

string_value:
	  '"'
	  { lexer_mode = LEXER_C_STRING; }
	  C_STRING
	  {lexer_mode = LEXER_NORMAL; }
	  '"'
	  {
	    $$ = new_string(buffer);
	    if ($$ == -1) YYERROR;
	  }
	;

function_call_value:
	  function_name '(' function_arguments ')'
	  {
	    char *s;
	    buffer_size = 0;
	    if (put('{') == -1) YYERROR;
	    s = get_string($1);
	    while (*s) { if (put(*s) == -1) YYERROR; s++; }
	    if (put('(') == -1) YYERROR;
	    s = get_string($3);
	    while (*s) { if (put(*s) == -1) YYERROR; s++; }
	    if (put(')') == -1) YYERROR;
	    if (put(';') == -1) YYERROR;
	    if (put('}') == -1) YYERROR;
	    if (put(0) == -1) YYERROR;
	    $$ = new_string(buffer);
	    if ($$ == -1) YYERROR;
	  }
	;

function_name:
	  IDENTIFIER
	  {
	    $$ = new_string(buffer);
	    if ($$ == -1) YYERROR;
	  }
	;

function_arguments:
	  { lexer_mode = LEXER_F_ARGS; }
	  F_ARGS
	  {
	    lexer_mode = LEXER_NORMAL;
	    $$ = new_string(buffer);
	    if ($$ == -1) YYERROR;
	  }
	;

c_code_value:
	  '{'
	  { lexer_mode = LEXER_C_CODE; }
	  C_CODE
	  { lexer_mode = LEXER_NORMAL; }
	  '}'
	  {
	    $$ = new_string(buffer);
	    if ($$ == -1) YYERROR;
	  }
	;

include_string:
	  { lexer_mode = LEXER_INCLUDE_STUFF; }
	  INCLUDE_STUFF
	  {
	    lexer_mode = LEXER_NORMAL;
	    $$ = new_string(buffer);
	    if ($$ == -1) YYERROR;
	  }

identifier:
	  IDENTIFIER
	  {
	    $$ = new_identifier(buffer);
	    if ($$ == -1) YYERROR;
	  }

keyword:
	  KEYWORD
	  {
	    $$ = buffer_keyword;
	  }
%%

/* C code */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* return 0 if ok, -1 if error */
int put(int c)
{
  if (buffer_size == buffer_maxsize) {
    char *nb;
    if (!buffer_size)
      nb = malloc(256);
    else
      nb = realloc(buffer, buffer_maxsize + 256);
    if (!nb) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    buffer = nb;
    buffer_maxsize += 256;
  }
  buffer[buffer_size] = c;
  buffer_size++;
  return 0;
}

int get_byte(void)
{
  int c = get();
  if (c == EOF) return -1;
  if (c == '\n') line++;
  return c;
}

void unget_byte(int c)
{
  if (c == '\n') line--;
  unget(c);
}

/* -1 error, 0 no comment, 1 comment removed */
int check_comments(int c, int put_it)
{
  if (c != '/') return 0;
  c = get_byte();

  if (c == '*') {
    if (put_it) if (put('/') == -1 || put('*') == -1) return -1;
    while (1) {
      do {
        c = get_byte();
        if (c == -1) { fprintf(stderr, "bad C comment\n"); return -1; }
        if (put_it) if (put(c) == -1) return -1;
      } while (c != '*');
      c = get_byte();
      if (c == '/') {
        if (put_it) if (put(c) == -1) return -1;
        return 1;
      }
      unget_byte(c);
    }
  }

  if (c == '/') {
    if (put_it) if (put('/') == -1 || put('/') == -1) return -1;
    while (1) {
      c = get_byte();
      if (put_it && c != -1) if (put(c) == -1) return -1;
      if (c == '\n') return 1;
      if (c == -1) { fprintf(stderr, "bad C comment\n"); return -1; }
    }
  }

  unget_byte(c);
  return 0;
}

/* return 0 if ok, -1 if error */
int get_full_string(int sep)
{
  int c = get_byte();
  int last_escape = 0;

  if (c != sep) return -1;

  do {
    if (put(c) == -1) return -1;
    if (last_escape) last_escape = 0; else last_escape = (c == '\\');
    c = get_byte();
    if (c == -1) {
      fprintf(stderr, "unterminated string sequence\n");
      return -1;
    }
  } while (c != sep || last_escape);

  if (put(sep) == -1) return -1;

  return 0;
}

static int _get_c_code(int put_it)
{
  int depth = 0;
  if (put_it) if (put('{') == -1) return -1;
  while (1) {
    int c = get_byte();
    int comment;

    if (c == -1) {
      fprintf(stderr,
              put_it ? "bad C code\n" : "bad init variables' block\n");
      return -1;
    }

    comment = check_comments(c, 1);
    if (comment == -1) return -1;
    if (comment == 1) continue;

    if (c == '}' && depth == 0) {
      unget_byte(c);
      if (put_it) if (put('}') == -1) return -1;
      if (put(0) == -1) return -1;
      if (put_it) return C_CODE;
      return INIT_VARS;
    }

    if (put(c) == -1) return -1;

    if (c == '{') depth++;
    if (c == '}') depth--;

    if (c == '"' || c == '\'') {
      buffer_size--;
      unget_byte(c);
      if (get_full_string(c) == -1)
        return -1;
    }
  }
}

int get_c_code(void)
{
  return _get_c_code(1);
}

int get_init_vars(void)
{
  return _get_c_code(0);
}

int get_c_code_no_bracket(void)
{
  /* let's use _get_c_code(0), but change its return value */
  int r = _get_c_code(0);
  if (r != INIT_VARS) return r;
  return C_CODE;
}

int get_c_string(void)
{
  int c = get_byte();
  int last_escape = 0;

  while (c != '"' || last_escape) {
    if (put(c) == -1) return -1;
    if (last_escape) last_escape = 0; else last_escape = (c == '\\');
    c = get_byte();
    if (c == -1) {
      fprintf(stderr, "bad C string\n");
      return -1;
    }
  }

  if (put(0) == -1) return -1;

  unget_byte('"');

  return C_STRING;
}

int get_include_stuff(void)
{
  while (1) {
    int c = get_byte();
    if (c == -1 || c == '\n') {
      if (put(0) == -1) return -1;
      break;
    }
    if (put(c) == -1) return -1;
  }

  return INCLUDE_STUFF;
}

int get_f_args(void)
{
  int c = get_byte();
  int depth = 0;

  while (c != ')' || depth) {
    int comment = check_comments(c, 0);
    if (comment == -1) return -1;

    if (!comment) if (put(c) == -1) return -1;
    if (c == '(') depth++;
    if (c == ')') depth--;
    if (c == '"' || c == '\'') {
      buffer_size--;
      unget_byte(c);
      if (get_full_string(c) == -1)
        return -1;
    }
    c = get_byte();
    if (c == -1) {
      fprintf(stderr, "bad function arguments\n");
      return -1;
    }
  }

  if (put(0) == -1) return -1;

  unget_byte(')');

  return F_ARGS;
}

int yylex(void)
{
  int c;
  int comment;
  keyword k;

lex_more:
  buffer_size = 0;

  if (lexer_mode == LEXER_C_CODE)
    return get_c_code();

  if (lexer_mode == LEXER_INIT_VARS)
    return get_init_vars();

  if (lexer_mode == LEXER_C_CODE_NO_BRACKET)
    return get_c_code_no_bracket();

  if (lexer_mode == LEXER_C_STRING)
    return get_c_string();

  if (lexer_mode == LEXER_F_ARGS)
    return get_f_args();

  if (lexer_mode == LEXER_INCLUDE_STUFF)
    return get_include_stuff();

  /* let's remove spaces */
  while (1) {
    c = get_byte();
    if (c == -1) return -1;
    if (!isspace(c))
      break;
  }

  /* keyword or identifier */
  if ((c >= 'a' && c <= 'z') ||
      (c >= 'A' && c <= 'Z') ||
      c == '_') {
    while (1) {
      if (put(c) == -1) return -1;
      c = get_byte();
      if ((c >= 'a' && c <= 'z') ||
          (c >= 'A' && c <= 'Z') ||
          (c >= '0' && c <= '9') ||
          c == '_')
        continue;
      unget_byte(c);
      if (put(0) == -1) return -1;
      switch ((k=lookup(buffer, buffer_size-1))) {
        case NULL_KEYWORD:
          return IDENTIFIER;
        case KEYWORD_WIDGET:
          return WIDGET;
        case KEYWORD_INCLUDE:
          return INCLUDE;
        case KEYWORD_CINCLUDE:
          return CINCLUDE;
        case KEYWORD_HINCLUDE:
          return HINCLUDE;
        case KEYWORD_INIT:
          if (init_first_shot == 0)
            return INIT;
          else
            return POST_INIT;
        case KEYWORD_CCODE:
          return CCODE;
        case KEYWORD_HCODE:
          return HCODE;
        default:
          buffer_keyword = k;
          return KEYWORD;
      }
    }
  }

  /* number (integer or float) */
  if (c >= '0' && c <= '9') {
    int dot_done = 0;
    while (1) {
      if (put(c) == -1) return -1;
      c = get_byte();
      if (c >= '0' && c <= '9')
        continue;
      if (c == '.' && !dot_done) {
        dot_done = 1;
        continue;
      }
      unget_byte(c);
      if (put(0) == -1) return -1;
      return dot_done ? FLOAT : INTEGER;
    }
  }

  /* comments */
  comment = check_comments(c, 0);
  if (comment == -1) return -1;
  if (comment == 1) goto lex_more;

  return c;
}

void yyerror(char *s)
{
  fprintf(stderr, "%s:%d:%s\n", get_current_file(), line+1, s);
}

void end_parse(void)
{
  if (!buffer_maxsize) return;
  free(buffer);
  buffer_size = buffer_maxsize = 0;
  buffer = 0;
}
