/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include <stdio.h>
#include <string.h>

#include "parser.h"
#include "string.h"
#include "widget.h"
#include "c_fun_args.h"
#include "files.h"
#include "generate.h"
#include "translate.h"
#include "gtk_widget.h"
#include "convert.h"
#include "globals.h"

/* this code is full of barbarian globals, this is bad taste but I live
 * in France.
 */
list cinclude;
list hinclude;
list ccode;
list hcode;

static char *long_documentation[] = {
#include "doc.h"
0
};

static char *short_documentation[] = {
#include "quick_doc.h"
0
};

static void show_doc(char **doc)
{
  while (*doc) fprintf(stdout, "%s\n", *doc++);
}

static void long_doc(void)
{
  show_doc(long_documentation);
}

static void short_doc(void)
{
  show_doc(short_documentation);
}

static void usage(void)
{
  fprintf(stderr, "usage:\n");
  fprintf(stderr,
      "    gtkcooker [options] <input file (extension has to be .k)>\n");
  fprintf(stderr, "options:\n");
  fprintf(stderr, "    -h | --help\n");
  fprintf(stderr, "        show this help message\n");
  fprintf(stderr, "    -d\n");
  fprintf(stderr, "        show long documentation\n");
  fprintf(stderr, "    -q\n");
  fprintf(stderr, "        show quick documentation\n");
  fprintf(stderr, "    -c <input encoding>\n");
  fprintf(stderr, "        convert the input file from <input encoding>\n");
  fprintf(stderr, "        (iso-8859-1 for example) to utf-8\n");
  fprintf(stderr, "    -l\n");
  fprintf(stderr, "        to have elements of widgets local by default\n");
  fprintf(stderr, "    -a\n");
  fprintf(stderr, "        to hide gtk types (void * in structs)\n");
  fprintf(stderr, "    -hout <file>\n");
  fprintf(stderr, "        to use <file> to put the .h stuff\n");
  fprintf(stderr, "http://www.nongnu.org/gtkcooker\n");
  fprintf(stderr, "This program is in the public domain.\n");
}

int main(int n, char **v)
{
  int ret = 0;
  char *input_file = 0;
  char *encoding = 0;
  int i;

  /* parsing command line arguments */
  for (i=1; i<n; i++) {
    if (!strcmp(v[i], "-c")) {
      /* one and only one encoding allowed */
      if (i == n-1 || encoding) {
        usage();
        goto error;
      }
      i++;
      encoding = v[i];
      continue;
    }
    if (!strcmp(v[i], "-l")) {
      default_local = 1;
      continue;
    }
    if (!strcmp(v[i], "-a")) {
      hide_gtk = 1;
      continue;
    }
    if (!strcmp(v[i], "-hout")) {
      /* one and only one hout allowed */
      if (i == n-1 || hout) {
        usage();
        goto error;
      }
      i++;
      hout = v[i];
      continue;
    }
    if (!strcmp(v[i], "-h") || !strcmp(v[i], "--help")) {
      usage();
      goto end;
    }
    if (!strcmp(v[i], "-d")) {
      long_doc();
      goto end;
    }
    if (!strcmp(v[i], "-q")) {
      short_doc();
      goto end;
    }
    /* the rest is not supported */
    if (v[i][0] == '-') {
      fprintf(stderr, "error: unsupported option %s\n", v[i]);
      goto error;
    }
    /* not an option => the file name (only one allowed) */
    if (input_file) {
      usage();
      goto error;
    }
    input_file = v[i];
  }

  if (!input_file) {
    usage();
    goto error;
  }

  /* let's go */
  if (encoding) if (init_convert(encoding) == -1) goto error;

  cinclude = new_list();
  hinclude = new_list();
  if (cinclude == -1 || hinclude == -1) goto error;

  ccode = new_list();
  hcode = new_list();
  if (ccode == -1 || hcode == -1) goto error;

  if (init_files(input_file) == -1) goto error;

  ret = yyparse();
  if (ret == 1) fprintf(stderr, "my god! parse error!\n");
  end_parse();
  if (ret == 1) goto error;

  if (translate() == -1) goto error;

  if (generate(input_file) == -1) goto error;

end:
  close_string();
  close_list();
  close_widget();
  close_arguments();
  close_identifier();
  close_gtk_widget();
  if (close_files() == -1) {
    fprintf(stderr, "error while closing a file, maybe the result of\n");
    fprintf(stderr, "gtkcooker won't be correct. Try to rerun it.\n");
    fprintf(stderr, "(Maybe you don't have enough space on your disk,\n");
    fprintf(stderr, "check that).\n");
    ret = 1;
  }
  close_convert();

  return ret;

error:
  ret = 1;
  goto end;
}
