/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include "translate.h"

#include <string.h>
#include <stdio.h>

#include "gtk_widget.h"
#include "widget.h"
#include "keywords.h"

extern widget_driver scroll_widget_driver;
extern widget_driver viewport_widget_driver;

/* -1 error, 0 ok */
static int translate_widget(widget w)
{
  int i;
  int n;
  widget_driver *driver;
  gtk_widget new;
  gtk_widget new2;

  n = widget_nb_elements(w);

  /* setup widget handlers */
  for (i=0; i<n; i++) {
    driver = find_driver(get_keyword(widget_element_type(w, i)));
    if (!driver || !driver->first_pass_translation) {
      fprintf(stderr, "unknown widget type: %s\n",
              get_keyword(widget_element_type(w, i)));
      return -1;
    }
    widget_element_set_driver(w, i, driver);
  }

  /* first pass of translation */
  for (i=0; i<n; i++) {
    driver = widget_element_driver(w, i);
    if (driver->first_pass_translation)
      if (driver->first_pass_translation(w, i) == -1) return -1;
  }

  /* second pass of translation */
  for (i=0; i<n; i++) {
    driver = widget_element_driver(w, i);
    if (driver->second_pass_translation)
      if (driver->second_pass_translation(w, i) == -1) return -1;
  }

  /* set recursive properties */
  n = widget_nb_gtk_widgets(w);
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    widget_driver *driver = WIDGET_DRIVER(gw);
    if (driver->populate_properties)
      if (driver->populate_properties(gw) == -1) return -1;
  }

  /* handle the scroll stuff - if a gtk_widget has the scroll property set,
   * then we must create a new gtk_widget that will impersonate the old one
   * (like a virus) (this is a hacker's project you know...)
   */
  n = widget_nb_gtk_widgets(w);
  for (i=0; i<n; i++) {
    gtk_widget gw = widget_get_gtk_widget(w, i);
    int type;
    if (!IS_SCROLL(gw)) continue;
    type = GET_TYPE(WIDGET_TYPE(gw));
    new = new_gtk_widget();
    if (new == -1) return -1;
    if (widget_add_gtk_widget(w, new) == -1) return -1;
    gtk_widget_copy(gw, new);
    /* let's impersonate gw (virus style!) */
    /* destroy the type, but not the properties */
    WIDGET_TYPE(gw) &= ~0xff;
    /* but force it to be local */
    SET_GLOBAL(gw, 0);
    WIDGET_TYPE(gw) |= SCROLL;
    WIDGET_NAME(gw) = new_name();
    if (WIDGET_NAME(gw) == -1) return -1;
    WIDGET_DRIVER(gw) = &scroll_widget_driver;
    SCROLL_CHILD(gw) = new;
    /* we must remove signals (ie. put a free list instead) */
    WIDGET_SIGNALS(gw) = new_list();
    if (WIDGET_SIGNALS(gw) == -1) return -1;

    /* maybe we must introduce a VIEWPORT */
    /* if you add a new widget, you may have to change this test so that
     * a VIEWPORT is not added if it has not to
     */
    if ((type == TEXT && TEXT_TYPE(new) == TEXT_MULTILINES) ||
        type == LIST)
      continue;
    new2 = new_gtk_widget();
    if (new2 == -1) return -1;
    if (widget_add_gtk_widget(w, new2) == -1) return -1;
    gtk_widget_copy(new, new2);
    /* let's impersonate new (virus style!) */
    /* destroy the type, but not the properties */
    WIDGET_TYPE(new) &= ~0xff;
    /* but force it to be local */
    SET_GLOBAL(new, 0);
    WIDGET_TYPE(new) |= VIEWPORT;
    WIDGET_NAME(new) = new_name();
    if (WIDGET_NAME(new) == -1) return -1;
    WIDGET_DRIVER(new) = &viewport_widget_driver;
    VIEWPORT_CHILD(new) = new2;
    /* we must remove signals (ie. put a free list instead) */
    WIDGET_SIGNALS(new) = new_list();
    if (WIDGET_SIGNALS(new) == -1) return -1;
  }

  return 0;
}

/* -1 error, 0 ok */
int translate(void)
{
  int i;
  int n;

  n = nb_widgets();
  for (i=0; i<n; i++)
    if (translate_widget(i) == -1) return -1;

  return 0;
}
