/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include "widget.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "gtk_widget.h"
#include "keywords.h"

static _widget *buffer;
static int size;
static int maxsize;

/* -1 error, widget if ok */
int new_widget(identifier name, list arguments)
{
  if (size == maxsize) {
    _widget *new;
    int new_size = maxsize + 32;
    if (!maxsize)
      new = malloc(new_size * sizeof(_widget));
    else
      new = realloc(buffer, new_size * sizeof(_widget));
    if (!new) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    buffer = new;
    maxsize = new_size;
  }

  buffer[size].name = name;
  buffer[size].arguments = arguments;
  buffer[size].vars = -1;
  buffer[size].init_code = -1;
  buffer[size].post_init_code = -1;
  buffer[size].data = 0;
  buffer[size].size = buffer[size].maxsize = 0;
  buffer[size].gtk_widgets = new_list();
  if (buffer[size].gtk_widgets == -1) return -1;

  size++;
  return size-1;
}

/* -1 error, widget_element if ok */
widget_element widget_find_element(widget wn, identifier n)
{
  _widget *w;
  int i;

  {
    static int rew=0;
    if (!rew) {
      fprintf(stderr, "%s:%d:%s: rewrite, too slow\n",
              __FILE__, __LINE__, __FUNCTION__);
      rew=1;
    }
  }

  w = &buffer[wn];

  for (i = 0; i < w->size; i++)
    if (!strcmp(get_identifier(n), get_identifier(w->data[i].name)))
      break;
  if (i == w->size) return -1;

  return i;
}

int widget_nb_elements(widget wn)
{
  return buffer[wn].size;
}

/* -1 error, 0 ok */
int widget_add_element(widget wn, keyword type, identifier name,
                       list arguments)
{
  _widget *w = &buffer[wn];
  int i;
  list l;

  if (w->size == w->maxsize) {
    _widget_element *new;
    int new_size = w->maxsize + 64;
    if (!maxsize)
      new = malloc(new_size * sizeof(_widget_element));
    else
      new = realloc(w->data, new_size * sizeof(_widget_element));
    if (!new) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    w->maxsize = new_size;
    w->data = new;
  }

  w->data[w->size].type = type;
  w->data[w->size].name = name;
  w->data[w->size].arguments = arguments;
  w->data[w->size].properties = new_list();
  if (w->data[w->size].properties == -1) return -1;
  w->data[w->size].pack = new_list();
  if (w->data[w->size].pack == -1) return -1;
  w->data[w->size].arguments_properties = new_list();
  if (w->data[w->size].arguments_properties == -1) return -1;
  for (i=0; i<list_size(arguments); i++) {
    l = new_list();
    if (l == -1) return -1;
    if (list_add_int(w->data[w->size].arguments_properties, l) == -1)
      return -1;
  }
  w->data[w->size].gtk_widget = -1;
  w->data[w->size].dad = -1;

  w->size++;

  return 0;
}

/* -1 error, 0 ok */
int set_init_vars(widget wn, char *b)
{
  _widget *w = &buffer[wn];
  string s = new_string(b);
  if (s == -1) return -1;
  w->vars = s;
  return 0;
}

/* -1 error, 0 ok */
int set_init_code(widget wn, string s)
{
  _widget *w = &buffer[wn];
  w->init_code = s;
  return 0;
}

/* -1 error, 0 ok */
int set_post_init_code(widget wn, string s)
{
  _widget *w = &buffer[wn];
  w->post_init_code = s;
  return 0;
}

/* -1 error, 0 ok */
int set_property(widget wn, widget_element we, keyword p, property_value v)
{
  _widget_element *w = &buffer[wn].data[we];
  if (list_add_int(w->properties, p) == -1 ||
      list_add_int(w->properties, v.type) == -1)
    return -1;
  if (v.type == PV_NUMBER) {
    if (list_add_float(w->properties, v.v.d) == -1)
      return -1;
  } else {
    if (list_add_int(w->properties, v.v.value) == -1)
      return -1;
  }
  return 0;
}

/* -1 error, 0 ok */
int set_argument_property(widget wn, widget_element we, int son,
                          keyword p, property_value v)
{
  _widget_element *w = &buffer[wn].data[we];
  list l = (list)list_get_int(w->arguments_properties, son);
  if (list_add_int(l, p) == -1 ||
      list_add_int(l, v.type) == -1)
    return -1;
  if (v.type == PV_NUMBER) {
    if (list_add_float(l, v.v.d) == -1)
      return -1;
  } else {
    if (list_add_int(l, v.v.value) == -1)
      return -1;
  }
  return 0;
}

/* -1 error, 0 ok */
int pack_new_row(widget wn, widget_element we, list row_attributes)
{
  _widget_element *w = &buffer[wn].data[we];
  list l = new_list();

  if (l == -1) return -1;

  if (list_add_int(l, row_attributes) == -1) return -1;

  if (list_add_int(w->pack, l) == -1) return -1;

  return 0;
}

/* -1 error, 0 ok */
int pack_append(widget wn, widget_element we, list add)
{
  _widget_element *w = &buffer[wn].data[we];
  list l;
  int i;

  /* if someone uses + before any >, it's an error */
  if (list_size(w->pack) == 0) {
    fprintf(stderr, "packing with '+' is not valid if '>' not used first\n");
    return -1;
  }

  l = (list)list_get_int(w->pack, list_size(w->pack) - 1);

  for (i=0; i<list_size(add); i++)
    if (list_add_int(l, list_get_int(add, i)) == -1)
      return -1;

  return 0;
}

/* -1 error, 0 ok */
int add_widget(widget wn, gtk_widget s)
{
  _widget *w = &buffer[wn];
  return list_add_int(w->gtk_widgets, s);
}

static void free_widget(_widget *w)
{
  if (!w->maxsize) return;

  free(w->data);
  w->data = 0;
  w->size = w->maxsize = 0;
}

void close_widget(void)
{
  int i;

  for (i=0; i<size; i++)
    free_widget(&buffer[i]);

  if (!maxsize) return;

  free(buffer);
  buffer = 0;
  size = maxsize = 0;
}

int nb_widgets(void)
{
  return size;
}

char *widget_name(widget wn)
{
  return get_identifier(buffer[wn].name);
}

char *widget_vars(widget wn)
{
  if (buffer[wn].vars == -1) return 0;
  return get_string(buffer[wn].vars);
}

list widget_arguments(widget wn)
{
  return buffer[wn].arguments;
}

/* -1 error, 0 ok */
int widget_add_gtk_widget(widget wn, gtk_widget e)
{
  _widget *w = &buffer[wn];
  return list_add_int(w->gtk_widgets, e);
}

int widget_nb_gtk_widgets(widget wn)
{
  return list_size(buffer[wn].gtk_widgets);
}

int widget_get_gtk_widget(widget wn, int i)
{
  return (gtk_widget)list_get_int(buffer[wn].gtk_widgets, i);
}

char *widget_init_code(widget wn)
{
  if (buffer[wn].init_code == -1) return 0;
  return get_string(buffer[wn].init_code);
}

char *widget_post_init_code(widget wn)
{
  if (buffer[wn].post_init_code == -1) return 0;
  return get_string(buffer[wn].post_init_code);
}

keyword widget_element_type(widget wn, widget_element we)
{
  return buffer[wn].data[we].type;
}

identifier widget_element_name(widget wn, widget_element we)
{
  return buffer[wn].data[we].name;
}

identifier widget_element_dad(widget wn, widget_element we)
{
  return buffer[wn].data[we].dad;
}

widget_driver *widget_element_driver(widget wn, widget_element we)
{
  return buffer[wn].data[we].driver;
}

list widget_element_arguments(widget wn, widget_element we)
{
  return buffer[wn].data[we].arguments;
}

list widget_element_properties(widget wn, widget_element we)
{
  return buffer[wn].data[we].properties;
}

list widget_element_arguments_properties(widget wn, widget_element we)
{
  return buffer[wn].data[we].arguments_properties;
}

list widget_element_pack(widget wn, widget_element we)
{
  return buffer[wn].data[we].pack;
}

gtk_widget widget_element_gtk_widget(widget wn, widget_element we)
{
  return buffer[wn].data[we].gtk_widget;
}

void widget_element_set_gtk_widget(widget wn, widget_element we, gtk_widget e)
{
  buffer[wn].data[we].gtk_widget = e;
}

void widget_element_set_dad(widget wn, widget_element we, widget_element dad)
{
  buffer[wn].data[we].dad = dad;
}

void widget_element_set_driver(widget wn, widget_element we,
                               widget_driver *driver)
{
  buffer[wn].data[we].driver = driver;
}

/* drivers of widgets */
extern widget_driver box_widget_driver;
extern widget_driver button_widget_driver;
extern widget_driver draw_widget_driver;
extern widget_driver file_widget_driver;
extern widget_driver image_widget_driver;
extern widget_driver label_widget_driver;
extern widget_driver list_widget_driver;
extern widget_driver menu_widget_driver;
extern widget_driver menu_item_widget_driver;
extern widget_driver resizer_widget_driver;
extern widget_driver scroll_widget_driver;
extern widget_driver separator_widget_driver;
extern widget_driver text_widget_driver;
extern widget_driver viewport_widget_driver;
extern widget_driver window_widget_driver;

/* drivers must appear in alphabetical order */
widget_driver *widget_drivers[NB_WIDGETS] = {
  &box_widget_driver,
  &button_widget_driver,
  &draw_widget_driver,
  &file_widget_driver,
  &image_widget_driver,
  &label_widget_driver,
  &list_widget_driver,
  &menu_widget_driver,
  &menu_item_widget_driver,
  &resizer_widget_driver,
  &scroll_widget_driver,
  &separator_widget_driver,
  &text_widget_driver,
  &viewport_widget_driver,
  &window_widget_driver,
};

/* return 0 if not found, the driver if found */
widget_driver *find_driver(char *name)
{
  int i;

  {
    static int done=0;
    if (!done) {
      done = 1;
      fprintf(stderr, "%s: %d: %s: rewrite, too slow\n",
              __FILE__, __LINE__, __FUNCTION__);
    }
  }

  /* hell, the list of drivers is alphabetically sorted
   * we should do some dichotomia here, no?
   */
  for (i=0; i<NB_WIDGETS; i++)
    if (!strcmp(name, widget_drivers[i]->name))
      return widget_drivers[i];

  return 0;
}
