/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#ifndef _WIDGET_H_
#define _WIDGET_H_

#include "identifier.h"
#include "list.h"
#include "string.h"

typedef int widget;
typedef int widget_element;

#include "gtk_widget.h"
#include "widget_driver.h"
#include "keywords.h"

typedef struct {
  keyword type;
  identifier name;
  list arguments;		/* <type 1> <value 1> ... <type n> <value n> */
  list properties;		/* <prop name> <prop type> <prop value> ... */
  list pack;			/* list of list
				 * each list has the structure:
				 * <attributes list> <widget_element 1> ...
				 * <widget element n>
				 * <attributes list> may be -1 otherwise it is
				 * a list of structure:
				 * <attribute name 1> <attribute type 1>
				 * <attribute value 1> ... <up to m attributes>
				 * a value of -2 as <widget element> may
				 * appear for 'menu' widget, meaning there is
				 * no submenu for the nth child.
				 * a value of -3 as <widget element> may
				 * appear for any widget, meaning there is
				 * a "separator for the nth child.
				 * a value of -4 as <widget element> may
				 * appear for any widget, meaning there is
				 * a "resizer between children.
				 */
  list arguments_properties;	/* list of list, each sublist is a list
				 * of properties:
				 * <prop name> <prop type> <prop value> ...
				 */
  gtk_widget gtk_widget;
  widget_element dad;
  struct widget_driver *driver;
} _widget_element;

typedef struct {
  identifier name;
  list arguments;	/* <arg 1 name> ... <arg n name> <arg string> */
  string vars;
  string init_code;
  string post_init_code;
  _widget_element *data;
  int size;
  int maxsize;
  list gtk_widgets;
} _widget;

/* widget argument types */
#define WA_KEYWORD      0
#define WA_IDENTIFIER   1
#define WA_NUMBER       2
#define WA_STRING       3

typedef struct {
  int type;
  union {
    int value;	/* for string, keyword or identifier */
    float d;	/* for float value */
  } v;
} widget_argument_value;

/* property value types */
#define PV_KEYWORD      0
#define PV_IDENTIFIER   1
#define PV_NUMBER       2
#define PV_STRING       3
#define PV_C_CODE       4

typedef struct {
  int type;
  union {
    int value;	/* for KEYWORD, IDENTIFIER, STRING, C_CODE */
    float d;	/* for NUMBER */
  } v;
} property_value;

int new_widget(identifier, list);
widget_element widget_find_element(widget, identifier);
int widget_nb_elements(widget);
int widget_add_element(widget, keyword, identifier, list);
int set_property(widget, widget_element, keyword, property_value);
int set_argument_property(widget, widget_element, int, keyword,
                          property_value);
int set_init_code(widget, string);
int set_post_init_code(widget, string);
int set_init_vars(widget, char *);
int pack_new_row(widget, widget_element, list);
int pack_append(widget, widget_element, list);
void close_widget(void);
int nb_widgets(void);
char *widget_name(widget);
char *widget_vars(widget);
list widget_arguments(widget);
int widget_add_gtk_widget(widget, gtk_widget);
int widget_nb_gtk_widgets(widget);
gtk_widget widget_get_gtk_widget(widget, int);
char *widget_init_code(widget);
char *widget_post_init_code(widget);

keyword widget_element_type(widget, widget_element);
identifier widget_element_name(widget, widget_element);
widget_element widget_element_dad(widget, widget_element);
struct widget_driver *widget_element_driver(widget, widget_element);
list widget_element_arguments(widget, widget_element);
list widget_element_properties(widget, widget_element);
list widget_element_arguments_properties(widget, widget_element);
list widget_element_pack(widget, widget_element);
gtk_widget widget_element_gtk_widget(widget, widget_element);
void widget_element_set_gtk_widget(widget, widget_element, gtk_widget);
void widget_element_set_dad(widget, widget_element, widget_element);
void widget_element_set_driver(widget, widget_element,
    struct widget_driver *);

#endif /* _WIDGET_H_ */
