/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include <string.h>

#include "widget.h"
#include "generate.h"
#include "files.h"
#include "widget.h"
#include "keywords.h"
#include "code.h"
#include "globals.h"

extern widget_driver label_widget_driver;
extern widget_driver button_widget_driver;

/* -1 error, 0 ok */
static int first_pass_translation(widget w, widget_element e)
{
  gtk_widget button;
  gtk_widget son;
  list l;
  int n;
  int i;
  int son_is_image = 0;

  if (list_size(widget_element_pack(w, e))) {
    fprintf(stderr, "packed button not supported yet (button %s)\n",
            get_identifier(widget_element_name(w, e)));
    return -1;
  }

  l = widget_element_arguments(w, e);
  if (list_size(l) != 2 ||
      (list_get_int(l, 0) != WA_STRING &&
       list_get_int(l, 0) != WA_IDENTIFIER)) {
b_err:
    fprintf(stderr,
            "button %s not properly declared. You must do something like:\n",
            get_identifier(widget_element_name(w, e)));
    fprintf(stderr, "  button %s(\"button's text\");\n",
            get_identifier(widget_element_name(w, e)));
    fprintf(stderr, "or:\n");
    fprintf(stderr, "  button %s(<image>);\n",
            get_identifier(widget_element_name(w, e)));
    return -1;
  }

  if (list_get_int(l, 0) == WA_IDENTIFIER) {
    widget_element im = widget_find_element(w, list_get_int(l, 1));
    if (im == -1 || widget_element_type(w, im) != KEYWORD_IMAGE) goto b_err;
    son_is_image = 1;
  }

  button = new_gtk_widget();
  if (button == -1) return -1;
  if (!son_is_image) {
    son = new_gtk_widget();
    if (son == -1) return -1;
  }

  if (widget_add_gtk_widget(w, button) == -1)
    return -1;
  if (!son_is_image && widget_add_gtk_widget(w, son) == -1)
    return -1;

  widget_element_set_gtk_widget(w, e, button);

  WIDGET_TYPE(button) = BUTTON;
  WIDGET_NAME(button) = widget_element_name(w, e);
  WIDGET_DRIVER(button) = &button_widget_driver;
  if (!default_local)
    SET_GLOBAL(button, 1);
  BUTTON_TYPE(button) = BUTTON_NORMAL;
  if (son_is_image)
    BUTTON_CHILD(button) = -1; /* will be set in the second_pass_translation */
  else
    BUTTON_CHILD(button) = son;

  if (!son_is_image) {
    WIDGET_NAME(son) = new_name();
    if (WIDGET_NAME(son) == -1) return -1;
    WIDGET_TYPE(son) = LABEL;
    WIDGET_DRIVER(son) = &label_widget_driver;
    LABEL_TYPE(son) = LABEL_STRING;
    LABEL_TEXT(son) = list_get_int(l, 1);
    LABEL_HALIGN(son) = -1;
    LABEL_VALIGN(son) = -1;
  }

  /* properties */
  l = widget_element_properties(w, e);
  n = list_size(l);
  for (i=0; i<n; i+=3) {
    keyword id = (keyword)list_get_int(l, i);
    int type = list_get_int(l, i+1);

    switch (id) {
      CASE_PROPERTY_CLICKED(button)
      CASE_PROPERTY_BUTTON_TYPE(button)
      CASE_PROPERTY_BUTTON_GROUP(button)
      CASE_PROPERTY_ACTIVE(button)
      CASE_PROPERTY_WIDTH(button)
      CASE_PROPERTY_HEIGHT(button)
      CASE_PROPERTY_BORDER(button, l, i)
      CASE_PROPERTY_GROW(button, l, i)
      CASE_PROPERTY_PADDING(button, l, i)
      CASE_PROPERTY_GLOBAL(button)
      CASE_PROPERTY_FOREGROUND_AND_BACKGROUND(button)
      CASE_PROPERTY_SCROLL(button)
      CASE_PROPERTY_HSCROLL(button)
      CASE_PROPERTY_VSCROLL(button)
      CASE_PROPERTY_DEFAULT
    }
  }

  return 0;
}

/* -1 error, 0 ok */
static int second_pass_translation(widget w, widget_element e)
{
  gtk_widget button = widget_element_gtk_widget(w, e);

  /* if child is -1 then child is in fact an image */
  if (BUTTON_CHILD(button) == -1) {
    /* no test done, all is ok from 1st pass translation */
    list l = widget_element_arguments(w, e);
    widget_element im = widget_find_element(w, list_get_int(l, 1));
    BUTTON_CHILD(button) = widget_element_gtk_widget(w, im);
  }

  /* group points to the widget, we must set it to the gtk_widget */
  /* we can't set it at first pass, because the other button might not have
   * been parsed yet, thus its gtk_widget might not exist
   */
  if (BUTTON_GROUP(button) != -1) {
    gtk_widget group = widget_element_gtk_widget(w, BUTTON_GROUP(button));
    BUTTON_GROUP(button) = group;
    /* type is forced to BUTTON_RADIO for both buttons */
    BUTTON_TYPE(button) = BUTTON_RADIO;
    BUTTON_TYPE(group) = BUTTON_RADIO;
  }

  return 0;
}

/* -1 error, 0 ok */
static int populate_properties(gtk_widget gw)
{
  int has_to_populate = 0;
  gtk_widget sw = BUTTON_CHILD(gw);
  widget_driver *sd = WIDGET_DRIVER(sw);

  POPULATE_PROPERTIES

  if (has_to_populate)
    if (sd->populate_properties)
      if (sd->populate_properties(sw) == -1) return -1;

  return 0;
}

/* -1 error, 0 ok */
static int generate_struct(gtk_widget gw)
{
  if (!IS_GLOBAL(gw)) return 0;

  if (hide_gtk)
    PUT_H("  /* GtkWidget* */ void *");
  else
    PUT_H("  GtkWidget *");
  PUT_H(get_identifier(WIDGET_NAME(gw)));
  PUT_H(";\n");

  return 0;
}

/* -1 error, 0 ok */
static int generate_signal_handlers(widget w, gtk_widget gw)
{
  int i;
  list l = WIDGET_SIGNALS(gw);
  for (i=0; i<list_size(l); i+=3) {
    keyword id = (keyword)list_get_int(l, i);
    switch (id) {
      CASE_GENERATE_SIGNAL_HANDLER_BUTTON_CLICKED
      CASE_GENERATE_SIGNAL_HANDLER_DEFAULT
    }
  }
  return 0;
}

/* -1 error, 0 ok */
static int generate_variable_declaration(gtk_widget gw)
{
  if (IS_GLOBAL(gw)) return 0;

  PUT_C("  GtkWidget *");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C(";\n");

  return 0;
}

/* -1 error, 0 ok */
static int generate_variable_instanciation(gtk_widget gw)
{
  PUT_C("  ");
  if (IS_GLOBAL(gw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  if (BUTTON_TYPE(gw) == BUTTON_NORMAL)
    PUT_C(" = gtk_button_new();\n");
  else if (BUTTON_TYPE(gw) == BUTTON_TOGGLE)
    PUT_C(" = gtk_toggle_button_new();\n");
  else if (BUTTON_TYPE(gw) == BUTTON_CHECK)
    PUT_C(" = gtk_check_button_new();\n");
  else
    PUT_C(" = gtk_radio_button_new(0);\n");

  GENERATE_WIDTH_AND_HEIGHT
  GENERATE_BORDER_WIDTH
  GENERATE_FOREGROUND_AND_BACKGROUND

  return 0;
}

/* -1 error, 0 ok */
static int generate_pack(gtk_widget gw)
{
  gtk_widget sw = BUTTON_CHILD(gw);
  PUT_C("  gtk_container_add(GTK_CONTAINER(");
  if (IS_GLOBAL(gw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C("), ");
  if (hide_gtk)
    PUT_C("(GtkWidget *)");
  if (IS_GLOBAL(sw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(sw)));
  PUT_C(");\n");

  if (BUTTON_GROUP(gw) != -1) {
    gtk_widget sw = BUTTON_GROUP(gw);
    gtk_widget dum;
    do {
      dum = sw;
      sw = BUTTON_GROUP(sw);
    } while (sw != -1);
    sw = dum;
    PUT_C("  gtk_radio_button_set_group(GTK_RADIO_BUTTON(");
    if (IS_GLOBAL(gw)) PUT_C("this->");
    PUT_C(get_identifier(WIDGET_NAME(gw)));
    PUT_C("),\n      gtk_radio_button_get_group(GTK_RADIO_BUTTON(");
    if (IS_GLOBAL(sw)) PUT_C("this->");
    PUT_C(get_identifier(WIDGET_NAME(sw)));
    PUT_C(")));\n");
  }

  return 0;
}

/* -1 error, 0 ok */
static int generate_post_pack(gtk_widget gw)
{
  if (IS_ACTIVE(gw)) {
    /* active must not be set if button is normal */
    if (BUTTON_TYPE(gw) == BUTTON_NORMAL) {
      fprintf(stderr, "warning: property \"active\" illegal for widget %s\n",
              get_identifier(WIDGET_NAME(gw)));
      SET_ACTIVE(gw, 0);
      return 0;
    }

    PUT_C("  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(");
    if (IS_GLOBAL(gw)) PUT_C("this->");
    PUT_C(get_identifier(WIDGET_NAME(gw)));
    PUT_C("), TRUE);\n\n");
  }

  return 0;
}

/* -1 error, 0 ok */
static int generate_signals(gtk_widget gw)
{
  int i;
  list l = WIDGET_SIGNALS(gw);

  for (i=0; i<list_size(l); i+=3) {
    keyword id = (keyword)list_get_int(l, i);

    switch (id) {
      CASE_GENERATE_SIGNAL_BUTTON_CLICKED
      CASE_GENERATE_SIGNAL_DEFAULT
    }
  }
  if (list_size(l)) PUT_C("\n");

  return 0;
}

widget_driver button_widget_driver = {
  .name = "button",
  .first_pass_translation = first_pass_translation,
  .second_pass_translation = second_pass_translation,
  .populate_properties = populate_properties,
  .generate_struct = generate_struct,
  .generate_signal_handlers = generate_signal_handlers,
  .generate_variable_declaration = generate_variable_declaration,
  .generate_variable_instanciation = generate_variable_instanciation,
  .generate_pack = generate_pack,
  .generate_post_pack = generate_post_pack,
  .generate_signals = generate_signals
};
