/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include <string.h>

#include "widget.h"
#include "generate.h"
#include "files.h"
#include "widget.h"
#include "keywords.h"
#include "code.h"
#include "globals.h"

extern widget_driver list_widget_driver;

/* -1 error, 0 ok */
static int first_pass_translation(widget w, widget_element e)
{
  int i;
  int n;
  list l;
  gtk_widget lst;

  if (list_size(widget_element_pack(w, e))) {
    fprintf(stderr, "packing in list is illegal (list %s)\n",
            get_identifier(widget_element_name(w, e)));
    return -1;
  }

  l = widget_element_arguments(w, e);
  if (list_size(l) < 4) {
l_decl_err:
    fprintf(stderr,
            "list %s not properly declared. You must do something like:\n",
            get_identifier(widget_element_name(w, e)));
    fprintf(stderr, "  list %s(\"list name\", \"elem 1\", ..., \"elem n\");\n",
            get_identifier(widget_element_name(w, e)));
    return -1;
  }
  for (i = 0; i < list_size(l); i+=2)
    if (list_get_int(l, i) != WA_STRING)
      goto l_decl_err;

  lst = new_gtk_widget();
  if (lst == -1) return -1;

  if (widget_add_gtk_widget(w, lst) == -1) return -1;

  widget_element_set_gtk_widget(w, e, lst);

  WIDGET_TYPE(lst) = LIST;
  WIDGET_NAME(lst) = widget_element_name(w, e);
  WIDGET_DRIVER(lst) = &list_widget_driver;
  if (!default_local)
    SET_GLOBAL(lst, 1);

  l = widget_element_arguments(w, e);
  n = list_size(l);

  LIST_NAME(lst) = (string)list_get_int(l, 1);
  LIST_DATA(lst) = new_list();
  if (LIST_DATA(lst) == -1) return -1;

  for (i=2; i<n; i+=2)
    if (list_add_int(LIST_DATA(lst), (string)list_get_int(l, i+1)) == -1)
      return -1;

  /* properties */
  l = widget_element_properties(w, e);
  n = list_size(l);
  for (i=0; i<n; i+=3) {
    keyword id = (keyword)list_get_int(l, i);
    int type = list_get_int(l, i+1);

    switch (id) {
      CASE_PROPERTY_BORDER(lst, l, i)
      CASE_PROPERTY_WIDTH(lst)
      CASE_PROPERTY_HEIGHT(lst)
      CASE_PROPERTY_GROW(lst, l, i)
      CASE_PROPERTY_GLOBAL(lst)
      CASE_PROPERTY_FOREGROUND_AND_BACKGROUND(lst)
      CASE_PROPERTY_LIST_ACTIVATE(lst)
      CASE_PROPERTY_SCROLL(lst)
      CASE_PROPERTY_HSCROLL(lst)
      CASE_PROPERTY_VSCROLL(lst)
      CASE_PROPERTY_DEFAULT
    }
  }

  return 0;
}

/* -1 error, 0 ok */
static int generate_struct(gtk_widget gw)
{
  if (!IS_GLOBAL(gw)) return 0;

  if (hide_gtk)
    PUT_H("  /* GtkWidget* */ void *");
  else
    PUT_H("  GtkWidget *");
  PUT_H(get_identifier(WIDGET_NAME(gw)));
  PUT_H(";\n");

  return 0;
}

/* -1 error, 0 ok */
static int generate_signal_handlers(widget w, gtk_widget gw)
{
  int i;
  list l = WIDGET_SIGNALS(gw);
  for (i=0; i<list_size(l); i+=3) {
    keyword id = (keyword)list_get_int(l, i);
    switch (id) {
      CASE_GENERATE_SIGNAL_HANDLER_LIST_ACTIVATE
      CASE_GENERATE_SIGNAL_HANDLER_DEFAULT
    }
  }
  return 0;
}

/* -1 error, 0 ok */
static int generate_variable_declaration(gtk_widget gw)
{
  if (IS_GLOBAL(gw)) return 0;

  PUT_C("  GtkWidget *");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C(";\n");

  return 0;
}

/* -1 error, 0 ok */
static int generate_variable_instanciation(gtk_widget gw)
{
  int i;

  PUT_C("  {\n");
  PUT_C("  GtkTreeIter _iter;\n");
  PUT_C("  GtkListStore *_l = gtk_list_store_new(1, G_TYPE_STRING);\n");
  PUT_C("  GtkCellRenderer *_renderer = gtk_cell_renderer_text_new();\n");
  PUT_C("  GtkTreeViewColumn *_column =\n");
  PUT_C("    gtk_tree_view_column_new_with_attributes (\"");
  PUT_C(get_string((string)LIST_NAME(gw)));
  PUT_C("\",\n");
  PUT_C("      _renderer, \"text\", 0, NULL);\n");
  for (i=0; i<list_size(LIST_DATA(gw)); i++) {
    PUT_C("  gtk_list_store_append(_l, &_iter);\n");
    PUT_C("  gtk_list_store_set(_l, &_iter, 0, \"");
    PUT_C(get_string((string)list_get_int(LIST_DATA(gw), i)));
    PUT_C("\", -1);\n");
  }
  PUT_C("  ");
  if (IS_GLOBAL(gw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C(" = gtk_tree_view_new_with_model(GTK_TREE_MODEL(_l));\n");
  PUT_C("  gtk_tree_view_append_column (GTK_TREE_VIEW(");
  if (IS_GLOBAL(gw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C("), _column);\n");
  PUT_C("  gtk_tree_selection_set_mode(gtk_tree_view_get_selection");
  PUT_C("(GTK_TREE_VIEW(");
  if (IS_GLOBAL(gw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C(")),\n");
  PUT_C("                              GTK_SELECTION_BROWSE);\n");
  PUT_C("  }\n");

  GENERATE_WIDTH_AND_HEIGHT
  GENERATE_BORDER_WIDTH
  GENERATE_FOREGROUND_AND_BACKGROUND

  return 0;
}

/* -1 error, 0 ok */
static int generate_signals(gtk_widget gw)
{
  int i;
  list l = WIDGET_SIGNALS(gw);

  for (i=0; i<list_size(l); i+=3) {
    keyword id = (keyword)list_get_int(l, i);

    switch (id) {
      CASE_GENERATE_SIGNAL_LIST_ACTIVATE
      CASE_GENERATE_SIGNAL_DEFAULT
    }
  }
  if (list_size(l)) PUT_C("\n");

  return 0;
}

widget_driver list_widget_driver = {
  .name = "list",
  .first_pass_translation = first_pass_translation,
  .second_pass_translation = 0,
  .populate_properties = 0,
  .generate_struct = generate_struct,
  .generate_signal_handlers = generate_signal_handlers,
  .generate_variable_declaration = generate_variable_declaration,
  .generate_variable_instanciation = generate_variable_instanciation,
  .generate_pack = 0,
  .generate_post_pack = 0,
  .generate_signals = generate_signals
};
