/*
 * gtkcooker: an abstraction layer on top of gtk
 * http://www.nongnu.org/gtkcooker
 * Public domain.
 */

#include <string.h>

#include "widget.h"
#include "generate.h"
#include "files.h"
#include "widget.h"
#include "keywords.h"
#include "code.h"
#include "globals.h"

extern widget_driver menu_widget_driver;
extern widget_driver menu_item_widget_driver;
extern widget_driver label_widget_driver;

/* -1 error, 0 ok */
static int menu_item_properties(widget w, widget_element e, int son,
                                gtk_widget it, list l)
{
  int i;
  int n = list_size(l);

  for (i=0; i<n; i+=3) {
    keyword id = (keyword)list_get_int(l, i);
    int type = list_get_int(l, i+1);

    switch (id) {
      CASE_PROPERTY_MENU_ACTIVATE(it)
      CASE_PROPERTY_DEFAULT_ITEM
    }
  }

  return 0;
}

/* -1 error, 0 ok */
static int first_pass_translation(widget w, widget_element e)
{
  int i;
  int n;
  list l;
  gtk_widget menu;
  int type = MENU_MENU;
  char submenu_name[32];

  /* if dad exists and is not 'menu', we are a menubar */
  widget_element dad = widget_element_dad(w, e);
  if (dad != -1 && widget_element_type(w, dad) != KEYWORD_MENU)
    type = MENU_MENUBAR;

  menu = new_gtk_widget();
  if (menu == -1) return -1;

  if (widget_add_gtk_widget(w, menu) == -1) return -1;

  widget_element_set_gtk_widget(w, e, menu);

  WIDGET_TYPE(menu) = MENU;
  WIDGET_NAME(menu) = widget_element_name(w, e);
  WIDGET_DRIVER(menu) = &menu_widget_driver;
  if (!default_local)
    SET_GLOBAL(menu, 1);
  SET_EXPAND(menu, 1);
  SET_FILL(menu, 1);
  MENU_TYPE(menu) = type;
  MENU_CHILDREN(menu) = new_list();
  if (MENU_CHILDREN(menu) == -1) return -1;

  l = widget_element_arguments(w, e);
  n = list_size(l);
  for (i=0; i<n; i+=2) {
    int type = list_get_int(l, i);

    if (type == WA_STRING) {
      gtk_widget menu_item;
      gtk_widget label;
      string sn;

      menu_item = new_gtk_widget();
      label = new_gtk_widget();
      if (menu_item == -1 || label == -1) return -1;

      if (widget_add_gtk_widget(w, menu_item) == -1 ||
          widget_add_gtk_widget(w, label) == -1)
        return -1;

      WIDGET_TYPE(menu_item) = MENU_ITEM;
      /* the name is derived from the widget's name */
      sprintf(submenu_name, "_%d", i/2);
      sn = new_string_concat(get_identifier(widget_element_name(w, e)),
                             submenu_name);
      WIDGET_NAME(menu_item) = new_identifier(get_string(sn));
      if (WIDGET_NAME(menu_item) == -1) return -1;
      WIDGET_DRIVER(menu_item) = &menu_item_widget_driver;
      MENU_ITEM_CHILD(menu_item) = label;
      MENU_ITEM_SUBMENU(menu_item) = -1;
      if (menu_item_properties(w, e, i/2, menu_item, (list)list_get_int(
              widget_element_arguments_properties(w, e), i/2)) == -1)
        return -1;

      WIDGET_TYPE(label) = LABEL;
      WIDGET_NAME(label) = new_name();
      if (WIDGET_NAME(label) == -1) return -1;
      WIDGET_DRIVER(label) = &label_widget_driver;
      LABEL_TYPE(label) = LABEL_STRING;
      LABEL_TEXT(label) = (string)list_get_int(l, i+1);
      LABEL_HALIGN(label) = 0;
      LABEL_VALIGN(label) = -1;

      if (list_add_int(MENU_CHILDREN(menu), menu_item) == -1) return -1;
    } else

    if (type == WA_KEYWORD) {
      keyword k = (keyword)list_get_int(l, i+1);
      if (k != KEYWORD_SEPARATOR) {
        fprintf(stderr,
          "keyword %s not allowed in declaration of menu %s in widget %s\n",
          get_keyword(k), get_identifier(widget_element_name(w, e)),
          widget_name(w));
        return -1;
      }
      if (list_add_int(MENU_CHILDREN(menu), -2) == -1) return -1;
    } else {

      fprintf(stderr,
              "bad declaration of menu %s in widget %s\n",
              get_identifier(widget_element_name(w, e)), widget_name(w));
      return -1;
    }
  }

  /* properties */
  l = widget_element_properties(w, e);
  n = list_size(l);
  for (i=0; i<n; i+=3) {
    keyword id = (keyword)list_get_int(l, i);
    int type = list_get_int(l, i+1);

    switch (id) {
      CASE_PROPERTY_GLOBAL(menu)
      CASE_PROPERTY_SCROLL(menu)
      CASE_PROPERTY_HSCROLL(menu)
      CASE_PROPERTY_VSCROLL(menu)
      CASE_PROPERTY_GROW(menu, l, i);
      CASE_PROPERTY_DEFAULT
    }
  }

  return 0;
}

/* -1 error, 0 ok */
static int second_pass_translation(widget w, widget_element e)
{
  int i;
  list pack_list;
  int pack_n;
  gtk_widget e_widget;
  int e_list;
  int e_n;

  if (list_size(widget_element_pack(w, e)) == 0) return 0;

  if (list_size(widget_element_pack(w, e)) != 1) {
    fprintf(stderr,
            "packing menu %s in widget %s illegal, only one > is allowed\n",
            get_identifier(widget_element_name(w, e)), widget_name(w));
    return -1;
  }

  pack_list = (list)list_get_int(widget_element_pack(w, e), 0);
  pack_n = list_size(pack_list) - 1;

  e_widget = widget_element_gtk_widget(w, e);
  e_list = MENU_CHILDREN(e_widget);
  e_n = list_size(e_list);

  for (i=0; i<pack_n; i++) {
    /* first element of pack list is of no use here */
    widget_element menu = (widget_element)list_get_int(pack_list, i + 1);
    gtk_widget menu_widget;
    gtk_widget child;
    if (menu == -2) continue;
    if (menu == -3) {
      fprintf(stderr,
  "warning: using \"separator\" in the packing of a menu, treating as a 0\n");
      continue;
    }
    if (menu == -4) {
      fprintf(stderr,
  "warning: using \"resizer\" in the packing of a menu, treating as a 0\n");
      continue;
    }
    menu_widget = widget_element_gtk_widget(w, menu);
    if (menu_widget == -1) {
      fprintf(stderr, "untranslated widget_element? (contact sed@free.fr)\n");
      return -1;
    }
    if (GET_TYPE(WIDGET_TYPE(menu_widget)) != MENU) {
      fprintf(stderr,
          "packing element %s in element %s of widget %s is not allowed\n",
          get_identifier(widget_element_name(w, menu)),
          get_identifier(widget_element_name(w, e)),
          widget_name(w));
      fprintf(stderr, "only elements of type 'menu' are allowed\n");
      fprintf(stderr, "type=%d\n", GET_TYPE(WIDGET_TYPE(menu_widget))); 
      return -1;
    }
    if (i >= e_n) {
      fprintf(stderr,
          "packing %s in element %s of widget %s is not allowed\n",
          get_identifier(widget_element_name(w, menu)),
          get_identifier(widget_element_name(w, e)),
          widget_name(w));
      fprintf(stderr,
              "because it does not correspond to an item of the menu\n");
      return -1;
    }
    child = (gtk_widget)list_get_int(e_list, i);
    if (child == -2) {
      fprintf(stderr,
          "packing %s in element %s of widget %s is not allowed\n",
          get_identifier(widget_element_name(w, menu)),
          get_identifier(widget_element_name(w, e)),
          widget_name(w));
      fprintf(stderr,
              "because it is a separator of the menu\n");
      return -1;
    }
    /* we can't have -3 nor -4, -3 and -4 are in packing, not in args,
     * so not in children
     */
    MENU_ITEM_SUBMENU(child) = menu_widget;
  }

  return 0;
}

/* -1 error, 0 ok */
static int populate_properties(gtk_widget gw)
{
  int has_to_populate;
  list l = MENU_CHILDREN(gw);
  int n = list_size(l);
  int i;
  gtk_widget sw;
  widget_driver *sd;

  for (i=0; i<n; i++) {
    sw = (gtk_widget)list_get_int(l, i);
    if (sw == -2) continue;
    sd = WIDGET_DRIVER(sw);
    has_to_populate = 0;

    POPULATE_PROPERTIES

    if (has_to_populate)
      if (sd->populate_properties)
        if (sd->populate_properties(sw) == -1) return -1;
  }

  return 0;
}

/* -1 error, 0 ok */
static int generate_struct(gtk_widget gw)
{
  if (!IS_GLOBAL(gw)) return 0;

  if (hide_gtk)
    PUT_H("  /* GtkWidget* */ void *");
  else
    PUT_H("  GtkWidget *");
  PUT_H(get_identifier(WIDGET_NAME(gw)));
  PUT_H(";\n");

  return 0;
}

#if 0
/* -1 error, 0 ok */
static int generate_signal_handlers(widget w, gtk_widget gw)
{
  return 0;
}
#endif

/* -1 error, 0 ok */
static int generate_variable_declaration(gtk_widget gw)
{
  if (IS_GLOBAL(gw)) return 0;

  PUT_C("  GtkWidget *");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C(";\n");

  return 0;
}

/* -1 error, 0 ok */
static int generate_variable_instanciation(gtk_widget gw)
{
  PUT_C("  ");
  if (IS_GLOBAL(gw)) PUT_C("this->");
  PUT_C(get_identifier(WIDGET_NAME(gw)));
  PUT_C(" = gtk_menu_");
  if (MENU_TYPE(gw) == MENU_MENUBAR) PUT_C("bar_");
  PUT_C("new();\n");

  GENERATE_FOREGROUND_AND_BACKGROUND

  return 0;
}

/* -1 error, 0 ok */
static int generate_pack(gtk_widget gw)
{
  list l = MENU_CHILDREN(gw);
  int n = list_size(l);
  int j;
  gtk_widget sw;
  for (j=0; j<n; j++) {
    PUT_C("  gtk_menu_shell_append(GTK_MENU_SHELL(");
    if (IS_GLOBAL(gw)) PUT_C("this->");
    PUT_C(get_identifier(WIDGET_NAME(gw)));
    PUT_C("), ");
    sw = (gtk_widget)list_get_int(l, j);
    /* we can't have -3 nor -4, -3 and -4 are in packing, not in args,
     * so not in children
     */
    if (sw == -2) {
      PUT_C("\n      gtk_separator_menu_item_new()");
    } else {
      if (hide_gtk)
        PUT_C("(GtkWidget *)");
      if (IS_GLOBAL(sw)) PUT_C("this->");
      PUT_C(get_identifier(WIDGET_NAME(sw)));
    }
    PUT_C(");\n");
  }

  return 0;
}

#if 0
/* -1 error, 0 ok */
static int generate_signals(gtk_widget gw)
{
  return 0;
}
#endif

widget_driver menu_widget_driver = {
  .name = "menu",
  .first_pass_translation = first_pass_translation,
  .second_pass_translation = second_pass_translation,
  .populate_properties = populate_properties,
  .generate_struct = generate_struct,
  .generate_signal_handlers = 0,
  .generate_variable_declaration = generate_variable_declaration,
  .generate_variable_instanciation = generate_variable_instanciation,
  .generate_pack = generate_pack,
  .generate_post_pack = 0,
  .generate_signals = 0
};
